/* $Id: dns.C,v 1.9 1999/11/13 03:37:24 dm Exp $ */

/*
 *
 * Copyright (C) 1998 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */


/* Asynchronous resolver.  This code is easy to understand if and only
 * if you also have a copy of RFC 1035.  You can get this by anonymous
 * ftp from, among other places, ftp.internic.net:/rfc/rfc1035.txt.
 */

#include "dnsparse.h"
#include "arena.h"

#include "vector.h"
#include "ihash.h"
#include "list.h"
#include "backoff.h"

INITFN(dns_start);

#define NSPORT NAMESERVER_PORT
#define QBSIZE 512

class dnsreq {
  static u_int16_t lastid;	// Last request id
  static int sock;		// UDP socket connected to nameserver

  static int num_timeout;
  static time_t first_timeout;

  static void resend (dnsreq *r) {r->xmit ();}
  static u_int16_t genid ();

  dnsreq ();
  virtual void readreply (dnsparse *) = 0;

  void fail (int);
  void start (bool);

  int srchno;

protected:
  int error;

public:
  static void setsock (bool);
  static void pktready ();

  u_int16_t id;			// DNS query ID
  str basename;			// Name for which to search
  str name;			// Name for which to query
  u_short type;			// Type of query (T_A, T_PTR, etc.)

  ihash_entry<dnsreq> hlink;	// Per-id hash table link
  tmoq_entry<dnsreq> tlink;	// Retransmit queue link

  dnsreq (str, u_int16_t, bool search = false);
  virtual ~dnsreq ();

  bool xmit ();
  void timeout ();
};

u_int16_t dnsreq::lastid;
int dnsreq::sock = -1;
int dnsreq::num_timeout;
time_t dnsreq::first_timeout;

static ihash<u_short, dnsreq, &dnsreq::id, &dnsreq::hlink> reqtab;
static tmoq<dnsreq, &dnsreq::tlink, 1, 5> reqtoq;

u_int16_t
dnsreq::genid ()
{
#ifndef HAVE_ARC4RANDOM
  static u_short dns_lastid;
  return ++dns_lastid;
#else /* HAVE_ARC4RANDOM */
  return arc4random () & 0xffff;
#endif /* HAVE_ARC4RANDOM */
}

void
dnsreq::fail (int err)
{
  if (!error)
    error = err;
  readreply (NULL);
  delete this;
}

void
dnsreq::start (bool again)
{
  if (again && (srchno < 0 || !_res.dnsrch[srchno])) {
    fail (NXDOMAIN);
    return;
  }

  if (again) {
    reqtab.remove (this);
    reqtoq.remove (this);
  }
  if (srchno >= 0)
    name = strbuf ("%s.%s", basename.cstr (), _res.dnsrch[srchno++]);
  id = genid ();
  reqtab.insert (this);
  reqtoq.start (this);
}

dnsreq::dnsreq (str n, u_int16_t t, bool search)
  : error (0), type (t)
{
  if (!search || strchr (n, '.')) {
    srchno = -1;
    basename = NULL;
    name = n;
  }
  else {
    srchno = 0;
    basename = n;
    name = NULL;
  }
  start (false);
}

dnsreq::~dnsreq ()
{
  reqtab.remove (this);
  reqtoq.remove (this);
}

//* Send a query to the nameserver */
bool
dnsreq::xmit ()
{
  u_char qb[QBSIZE];
  HEADER *const h = (HEADER *) qb;
  int n;

  n = res_mkquery (QUERY, name, C_IN, type, NULL, 0, NULL, qb, sizeof (qb));
  if (n < 0)
    return false;
  h->id = htons (id);  /* XXX - Why bother with htons? */
  if (send (sock, reinterpret_cast<char *> (qb), n, 0) < 0) {
    error = ARERR_CANTSEND;
    return false;
  }
  return true;
}

void
dnsreq::timeout ()
{
  num_timeout++;
  if (!first_timeout)
    time (&first_timeout);
  if (num_timeout > 5 && time (NULL) - first_timeout > 64)
    setsock (true);
  fail (ARERR_TIMEOUT);
}

void
dnsreq::pktready ()
{
  u_char qb[QBSIZE];
  ssize_t n;

  while ((n = recv (sock, reinterpret_cast<char *> (qb),
		    sizeof (qb), 0)) >= 0) {
    num_timeout = 0;
    first_timeout = 0;

    dnsparse reply (qb, n);
    question q;
    if (!reply.qparse (&q))
      continue;

    dnsreq *r;
    for (r = reqtab[ntohs (reply.hdr->id)];
	 r && (r->type != q.q_type || strcasecmp (r->name, q.q_name));
	 r = reqtab.nextkeq (r))
      ;
    if (!r)
      continue;

    if (reply.error && !r->error)
      r->error = reply.error;
    if (r->error == NXDOMAIN) {
      r->error = 0;
      r->start (true);
    }
    else {
      r->readreply (r->error ? NULL : &reply);
      delete r;
    }
  }
  if (n < 0 && errno == ECONNREFUSED)
    setsock (true);
}

void
dnsreq::setsock (bool bump)
{
  static int ns_idx;		// Index of current server in nsaddr_list

  if (sock >= 0) {
    fdcb (sock, selread, NULL);
    close (sock);
  }
  /* We use a reserved port here to prevent port-capture problems.
   * (An unprivileged user can steal packets destined to an INADDR_ANY
   * socket by binding to the same port on the more-specific interface
   * address. */
  sock = inetsocket_resvport (SOCK_DGRAM, INADDR_ANY);
  if (sock < 0)
    fatal ("dnsreq::setsock: inetsocket: %s\n", strerror (errno));
  close_on_exec (sock);
  if (bump && _res.nscount) {
    ns_idx = (ns_idx + 1) % _res.nscount;
    warn ("changing nameserver to %s\n",
	  inet_ntoa (_res.nsaddr_list[ns_idx].sin_addr));
  }
  if (!_res.nsaddr_list[ns_idx].sin_addr.s_addr) /* XXX */
    _res.nsaddr_list[ns_idx].sin_addr.s_addr = htonl (INADDR_LOOPBACK);
  if (connect (sock, (struct sockaddr *) &_res.nsaddr_list[ns_idx],
	       sizeof (_res.nsaddr_list[0])) < 0)
    fatal ("dnsreq::setsock: connect: %s\n", strerror (errno));
  make_async (sock);
  fdcb (sock, selread, wrap (pktready));
  num_timeout = 0;
  first_timeout = 0;
  reqtab.traverse (wrap (resend));
}

class dnsreq_a : public dnsreq {
  bool checkaddr;		// Non-zero when arr_addr must be checked
  in_addr addr;			// Adress of inverse queries (for checking)
  cbhent cb;			// Callback for hostbyname/addr
  dnsreq_a ();
public:
  dnsreq_a (str n, cbhent c, bool s = false)
    : dnsreq (n, T_A, s), checkaddr (false), cb (c) {}
  dnsreq_a (str n, cbhent c, const in_addr &a)
    : dnsreq (n, T_A), checkaddr (true), addr (a), cb (c) {}
  void readreply (dnsparse *);
};

void
dnsreq_a::readreply (dnsparse *reply)
{
  ptr<hostent> h;
  if (!error) {
    if (!(h = reply->tohostent ()))
      error = reply->error;
    else if (checkaddr) {
      char **ap;
      for (ap = h->h_addr_list; *ap && *(in_addr *) *ap != addr; ap++)
	;
      if (!*ap) {
	h = NULL;
	error = ARERR_PTRSPOOF;
      }
    }
  }
  (*cb) (h, error);
}

void
dns_hostbyname (str name, cbhent cb,
		bool search, bool addrok)
{
  if (addrok) {
    in_addr addr;
    if (inet_aton (name.cstr (), &addr)) {
      ptr<hostent> h = refcounted<hostent, vsize>::alloc
	(sizeof (*h) + 3 * sizeof (void *)
	 + sizeof (addr) + strlen (name) + 1);
      h->h_aliases = (char **) &h[1];
      h->h_addrtype = AF_INET;
      h->h_length = sizeof (addr);
      h->h_addr_list = &h->h_aliases[1];

      h->h_aliases[0] = NULL;
      h->h_addr_list[0] = (char *) &h->h_addr_list[2];
      h->h_addr_list[1] = NULL;

      *(struct in_addr *) h->h_addr_list[0] = addr;
      h->h_name = (char *) h->h_addr_list[0] + sizeof (addr);
      strcpy ((char *) h->h_name, name);

      (*cb) (h, 0);
      return;
    }
  }
  vNew dnsreq_a (name, cb, search);
}

class dnsreq_ptr : public dnsreq {
  in_addr addr;
  cbhent cb;			// Callback for hostbyname/addr
  dnsreq_ptr ();
public:
  static str inaddr_arpa (in_addr);
  dnsreq_ptr (in_addr a, cbhent c)
    : dnsreq (inaddr_arpa (a), T_PTR), addr (a), cb (c) {}
  void readreply (dnsparse *);
};

str
dnsreq_ptr::inaddr_arpa (in_addr addr)
{
  u_char *a = reinterpret_cast<u_char *> (&addr);
  return strbuf ("%d.%d.%d.%d.in-addr.arpa", a[3], a[2], a[1], a[0]);
}

void
dnsreq_ptr::readreply (dnsparse *reply)
{
  if (!error) {
    const u_char *cp = reply->getanp ();
    for (u_int i = 0; i < reply->ancount; i++) {
      resrec rr;
      if (!reply->rrparse (&cp, &rr))
	break;
      if (rr.rr_type == T_PTR && rr.rr_class == C_IN) {
	vNew dnsreq_a (rr.rr_ptr, cb, addr);
	return;
      }
    }
  }
  if (!error && !(error = reply->error))
    error = ARERR_NXREC;
  (*cb) (NULL, error);
}

void
dns_hostbyaddr (in_addr addr, cbhent cb)
{
  vNew dnsreq_ptr (addr, cb);
}


static void
dns_start ()
{
  if (!(_res.options & RES_INIT))
    res_init ();
  dnsreq::setsock (false);
}

const char *
dns_strerror (int no)
{
  switch (no) {
  case NOERROR:
    return "no error";
  case FORMERR:
    return "DNS format error";
  case SERVFAIL:
    return "name server failure";
  case NXDOMAIN:
    return "non-existent domain name";
  case NOTIMP:
    return "unimplemented DNS request";
  case REFUSED:
    return "DNS query refused";
  case ARERR_NXREC:
    return "no DNS records of appropriate type";
  case ARERR_TIMEOUT:
    return "name lookup timed out";
  case ARERR_PTRSPOOF:
    return "incorrect PTR record";
  case ARERR_BADRESP:
    return "malformed DNS reply";
  case ARERR_CANTSEND:
    return "cannot send to name server";
  case ARERR_REQINVAL:
    return "malformed domain name";
  case ARERR_CNAMELOOP:
    return "CNAME records form loop";
  default:
    return "unknown DNS error";
  }
}

int
dns_tmperr (int no)
{
  switch (no) {
  case SERVFAIL:
  case ARERR_TIMEOUT:
  case ARERR_CANTSEND:
  case ARERR_BADRESP:
    return 1;
  default:
    return 0;
  }
}

